﻿//    Copyright (C) 2009  Regis COSNIER
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Net;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using OpenTK.Graphics;
using OpenTK.Math;
using System.Runtime.InteropServices;
using System.Text;
using System.Xml;
using System.Xml.XPath;

namespace StreetViewMapper
{
	public partial class MainForm : Form
	{
		enum DisplayMode { None, Panorama, MiniPlanet };
		DisplayMode displayMode = DisplayMode.None;

		string cacheFolder = "Cache";

		// Panorama
        Object3D[] panoramaSpheres;
		float panoramaTheta = 0.0f;
		float panoramaPhi = 0.0f;
		float panoramaZoom = 1.0f;
		int panoramaLoadingZoom = 3;
		bool debugMaterial = false;

		// MiniPlanet
		float miniPlanetZoom = 0.1f;
		float miniPlanetScale = 5.0f;
		float miniPlanetPower = 1.2f;
		float miniPlanetAngle = 0;
		Shader shader = new Shader();
		CubeMap cubeMap = new CubeMap();


		public MainForm()
		{
			InitializeComponent();

			glControl.MouseWheel += new MouseEventHandler(glControl_MouseWheel);

			textBoxStreetViewURL.Text = Properties.Settings.Default.URL;
			comboBoxZoom.SelectedIndex = Properties.Settings.Default.Zoom;
			comboBoxPlanetType.SelectedIndex = Properties.Settings.Default.PlanetType;
		}

		private void Form1_FormClosed(object sender, FormClosedEventArgs e)
		{
			Properties.Settings.Default.URL = textBoxStreetViewURL.Text;
			Properties.Settings.Default.Zoom = comboBoxZoom.SelectedIndex;
			Properties.Settings.Default.PlanetType = comboBoxPlanetType.SelectedIndex;
			Properties.Settings.Default.Save();
		}


		bool DebugMaterial
		{
			get { return debugMaterial; }
			set
			{
				if (debugMaterial != value)
				{
					debugMaterial = value;
					glControl.Invalidate();
				}
			}
		}

		Object3D[] CreateStreetViewSpheres(PanoramaParameters panoramaParameters, int loadingZoom)
		{
			int numThetaStep = 0;
			int numPhiStep = 0;
            double totalThetaUWidth = 0.0;
            double totalPhiVWidth = 0.0;

			// StreetViewAPI http://code.google.com/apis/maps/documentation/services.html
			// Google maps parameters http://mapki.com/wiki/Google_Map_Parameters
			// Hacking Google Street View http://wherecamp.pbworks.com/Hacking-Google-Street-View

			////http://cbk0.google.com/cbk?output=tile&panoid=9aoJaoXlaD08JWgy2NIwiQ&zoom=3&x=6&y=1&cb_client=maps_sv
			//switch (loadingZoom)
			//{
			//    case 0:
			//        // X: 360° -> 416px
			//        // Y: 180° -> 208px
			//        numThetaStep = 1;
			//        numPhiStep = 1;
			//        totalThetaUWidth = 416.0 / 512.0;
			//        totalPhiVWidth = 208.0 / 512.0;
			//        if (lasVegasAngleResolution)
			//        {
			//            // For Las Vegas: http://maps.google.com/maps?f=q&source=s_q&hl=en&geocode=&q=&ie=UTF8&t=h&layer=c&cbll=36.100924,-115.174624&panoid=9ki6pGBFN90vgcIFOzeQvQ&cbp=13,172.7227616306584,,0,3.012499999999994&ll=36.10092,-115.190792&spn=0,359.922752&z=14
			//            totalThetaUWidth = 300.0 / 512.0;
			//            totalPhiVWidth = 150.0 / 512.0;
			//        }
			//        break;
			//    case 1:
			//        // X: 360° -> 832px
			//        // Y: 180° -> 416px
			//        numThetaStep = 2;
			//        numPhiStep = 1;
			//        totalThetaUWidth = 832.0 / 512.0;
			//        totalPhiVWidth = 416.0 / 512.0;
			//        if (lasVegasAngleResolution)
			//        {
			//            totalThetaUWidth = 600.0 / 512.0;
			//            totalPhiVWidth = 300.0 / 512.0;
			//        }
			//        break;
			//    case 2:
			//        // X: 360° -> 1664px (1666px)
			//        // Y: 180° -> 832px
			//        numThetaStep = 4;
			//        numPhiStep = 2;
			//        totalThetaUWidth = 1664.0 / 512.0;
			//        totalPhiVWidth = 832.0 / 512.0;
			//        if (lasVegasAngleResolution)
			//        {
			//            numThetaStep = 3;
			//            totalThetaUWidth = 1200.0 / 512.0;
			//            totalPhiVWidth = 600.0 / 512.0;
			//        }
			//        break;
			//    case 3:
			//        // X: 360° -> 3328px
			//        // Y: 180° -> 1664px
			//        numThetaStep = 7;
			//        numPhiStep = 4;
			//        totalThetaUWidth = 3328.0 / 512.0;
			//        totalPhiVWidth = 1664.0 / 512.0;
			//        if (lasVegasAngleResolution)
			//        {
			//            numThetaStep = 5;
			//            numPhiStep = 3;
			//            totalThetaUWidth = 2400.0 / 512.0;
			//            totalPhiVWidth = 1200.0 / 512.0;
			//        }
			//        break;
			//    case 4:
			//        // X: 360° -> 6656px
			//        // Y: 180° -> 3328px
			//        numThetaStep = 13;
			//        numPhiStep = 7;
			//        totalThetaUWidth = 6656.0 / 512.0;
			//        totalPhiVWidth = 3328.0 / 512.0;
			//        break;
			//    case 5:
			//        // X: 360° -> 13312px
			//        // Y: 180° -> 6658px
			//        numThetaStep = 26;
			//        numPhiStep = 13;
			//        totalThetaUWidth = 13312.0 / 512.0;
			//        totalPhiVWidth = 6656.0 / 512.0;
			//        break;
			//}

			int zoomMax = 0;
			double searchMaxZoom = Math.Max((double)panoramaParameters.ImageWidth / (double)panoramaParameters.TileWidth, (double)panoramaParameters.ImageHeight / (double)panoramaParameters.TileHeight);
			do
			{
				zoomMax++;
				searchMaxZoom /= 2;
			}
			while ((int)Math.Ceiling(searchMaxZoom) > 1);

			if (loadingZoom > zoomMax)
				loadingZoom = zoomMax;
			int zoom = zoomMax - loadingZoom;
			double w = panoramaParameters.ImageWidth >> zoom;
			double h = panoramaParameters.ImageHeight >> zoom;
			numThetaStep = (int)Math.Ceiling(w / (double)panoramaParameters.TileWidth);
			numPhiStep = (int)Math.Ceiling(h / (double)panoramaParameters.TileHeight);
			totalThetaUWidth = w / 512.0;
			totalPhiVWidth = h / 512.0;



            List<Object3D> sphereParts = new List<Object3D>();

			int loadBalancingIndex = 0;

			double thetaStart = 0.0;
			double phiStart = -90.0;

            double thetaStep = 360.0 / totalThetaUWidth; // horizontal angle by tile
            double phiStep = 180.0 / totalPhiVWidth; // vertical angle by tile

            double theta = thetaStart;
            double thetaUWidth = totalThetaUWidth;
            double thetaRemaning = 360.0;
			for (int iTheta = 0; iTheta < numThetaStep; iTheta++)
			{
				double phi = phiStart;
                double phiVWidth = totalPhiVWidth;
                double phiRemaning = 180.0;
                for (int iPhi = 0; iPhi < numPhiStep; iPhi++)
				{
					string filename = string.Format("zoom={0}&x={1}&y={2}.jpg", loadingZoom, iTheta, iPhi);

					if (panoramaParameters.PanoramaId != null)
					{
						try
						{
							string url = string.Format("http://cbk{4}.google.com/cbk?output=tile&panoid={0}&zoom={1}&x={2}&y={3}&cb_client=maps_sv",
									panoramaParameters.PanoramaId, loadingZoom, iTheta, iPhi, loadBalancingIndex++ % 4);
							LogMessage(string.Format("Download {0}", url));
							WebRequest webRequest = WebRequest.Create(url);
							WebResponse webResponse = webRequest.GetResponse();
							Stream responseStream = webResponse.GetResponseStream();
							using (Stream fileStream = File.Open(Path.Combine(cacheFolder, filename), FileMode.Create))
							{
								int length = (int)webResponse.ContentLength;
								byte[] responseFile = new byte[length];
								int lengthRead = responseStream.Read(responseFile, 0, (int)length);
								while (lengthRead > 0)
								{
									fileStream.Write(responseFile, 0, lengthRead);
									lengthRead = responseStream.Read(responseFile, 0, (int)length);
								}
								fileStream.Close();
							}
							webResponse.Close();
						}
						catch (Exception ex)
						{
							throw new Exception(string.Format("Cannot load the image from Google StreetView server.\r\n{0}", ex.Message), ex);
						}
					}

					//string filename = string.Format("..\\..\\..\\Misc\\zoom{0}\\zoom={0}&x={1}&y={2}.jpg", zoom, iTheta, iPhi);
                    double thetaUW = Math.Min(1.0, thetaUWidth);
                    double phiVW = Math.Min(1.0, phiVWidth);
					Object3D spherePart = CreateStreetViewSphere(filename,
                        theta, theta + Math.Min(thetaRemaning, thetaStep),
                        phi, phi + Math.Min(phiRemaning, phiStep),
                        thetaUW, phiVW);
                    sphereParts.Add(spherePart);

                    phi += phiStep;
                    phiRemaning -= phiStep;
                    phiVWidth -= 1.0;
                }
				theta += thetaStep;
                thetaRemaning -= thetaStep;
                thetaUWidth -= 1.0;
			}

			return sphereParts.ToArray();
		}

        Object3D CreateStreetViewSphere(string textureFileName,
            double thetaStart, double thetaStop,
            double phiStart, double phiStop,
            double thetaUWidth, double phiVWidth)
        {
            int numThetaStep = 10;
            int numPhiStep = 10;

			LogMessage(string.Format("Load {0}", textureFileName));
            Material material = new Material();
			material.name = material.textureFileName = Path.Combine(cacheFolder, textureFileName);
			material.diffuseColor = new Vector4(
				(float)(thetaStart / 360.0),
				1.0f - (float)(thetaStart / 360.0),
				(float)((90.0 + phiStart) / 180.0), 1.0f);
			material.materialType = Material.MaterialType.Texture;
            material.Init();

            Object3D spherePart = new Object3D();
            spherePart.vertex = new float[3 * (numPhiStep + 1) * (numThetaStep + 1)];
            spherePart.texture = new float[2 * (numPhiStep + 1) * (numThetaStep + 1)];
            spherePart.indices = new int[6 * numPhiStep * numThetaStep];
            spherePart.faceType = Object3D.FaceType.Triangles;
            spherePart.material = material;

            int vertexIndex = 0;
            int textureIndex = 0;
            int indiceIndex = 0;

            double thetaStep = (thetaStop - thetaStart) / (double)numThetaStep;
            double phiStep = (phiStop - phiStart) / (double)numPhiStep;

            double thetaUStep = thetaUWidth / (double)numThetaStep;
            double phiVStep = phiVWidth / (double)numPhiStep;

            double thetaU = 0.0;

            double theta = thetaStart;
            for (int iTheta = 0; iTheta <= numThetaStep; iTheta++)
            {
                double phi = phiStart;
				double phiV = 0.0;
				for (int iPhi = 0; iPhi <= numPhiStep; iPhi++)
                {
                    double a = Math.Cos(-phi / 180.0 * Math.PI);
                    float z = (float)(a * Math.Cos(-theta / 180.0 * Math.PI));
                    float x = (float)(a * Math.Sin(-theta / 180.0 * Math.PI));
                    float y = (float)(Math.Sin(-phi / 180.0 * Math.PI));
                    spherePart.vertex[vertexIndex++] = x;
                    spherePart.vertex[vertexIndex++] = y;
                    spherePart.vertex[vertexIndex++] = z;

                    spherePart.texture[textureIndex++] = (float)thetaU;
                    spherePart.texture[textureIndex++] = (float)phiV;

                    if (iTheta < numThetaStep && iPhi < numPhiStep)
                    {
                        int i0 = (numPhiStep + 1) * (iTheta + 0) + (iPhi + 0);
                        int i1 = (numPhiStep + 1) * (iTheta + 0) + (iPhi + 1);
                        int i2 = (numPhiStep + 1) * (iTheta + 1) + (iPhi + 0);
                        int i3 = (numPhiStep + 1) * (iTheta + 1) + (iPhi + 1);

                        spherePart.indices[indiceIndex++] = i0;
                        spherePart.indices[indiceIndex++] = i1;
                        spherePart.indices[indiceIndex++] = i2;
                        spherePart.indices[indiceIndex++] = i2;
                        spherePart.indices[indiceIndex++] = i1;
                        spherePart.indices[indiceIndex++] = i3;
                    }

                    phi += phiStep;
                    phiV += phiVStep;
                }
                theta += thetaStep;
                thetaU += thetaUStep;
            }

            return spherePart;
        }

		void Form1_Load(object sender, EventArgs e)
		{
			ResizeControl();

			GL.ClearColor(Color.Gray);

			GL.Hint(HintTarget.PerspectiveCorrectionHint, HintMode.Nicest);
			//GL.Enable(EnableCap.CullFace);
			//GL.CullFace(CullFaceMode.Back);
			GL.Disable(EnableCap.Lighting);
		}

		void glControl_Paint(object sender, PaintEventArgs e)
		{
			glControl.MakeCurrent();

			if (displayMode == DisplayMode.Panorama)
			{
				GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);

				DrawSphere(panoramaPhi, panoramaTheta);
			}
			else if (displayMode == DisplayMode.MiniPlanet)
			{
				DrawMiniPlanet(glControl.Width, glControl.Height);
			}

			glControl.SwapBuffers();
		}

		private void panelScreen_Resize(object sender, EventArgs e)
		{
			System.Diagnostics.Trace.WriteLine(string.Format("{0}, {1}", Width, Height));
			ResizeControl();
		}

		private void ResizeControl()
		{
			int width = panelScreen.Size.Width;
			int height = panelScreen.Size.Height;
			if (width > 0 && height > 0)
			{
				if (displayMode == DisplayMode.MiniPlanet)
				{
					if (width > height)
					{
						glControl.Location = new Point((width - height) / 2, 0);
						glControl.Size = new System.Drawing.Size(height, height);
					}
					else
					{
						glControl.Location = new Point(0, (height - width) / 2);
						glControl.Size = new System.Drawing.Size(width, width);
					}
				}
				else
				{
					glControl.Location = new Point(0, 0);
					glControl.Size = new System.Drawing.Size(width, height);
				}
			}

			width = glControl.Size.Width;
			height = glControl.Size.Height;
			if (width > 0 && height > 0)
			{
				GL.Viewport(0, 0, width, height);

				GL.MatrixMode(MatrixMode.Projection);
				GL.LoadIdentity();

				if (displayMode == DisplayMode.Panorama)
					Glu.Perspective(90.0 * panoramaZoom, (double)width / (double)height, (double)0.1, (double)10.0);

				GL.MatrixMode(MatrixMode.Modelview);
				GL.LoadIdentity();
			}
			glControl.Invalidate();
		}

		void DrawSphere(float phi, float theta)
		{
			if (panoramaSpheres != null)
			{
				GL.LoadIdentity();
				//Glu.LookAt(Vector3.Zero, -Vector3.UnitZ, Vector3.UnitY);

				GL.Rotate(phi, Vector3.UnitX);
				GL.Rotate(theta, Vector3.UnitY);

				if (DebugMaterial)
				{
					GL.Enable(EnableCap.ColorMaterial);
					GL.TexEnv(TextureEnvTarget.TextureEnv, TextureEnvParameter.TextureEnvMode, (int)TextureEnvMode.Modulate);

					foreach (Object3D spherePart in panoramaSpheres)
					{
						GL.Enable(EnableCap.DepthTest);
						GL.PolygonOffset(1.0f, 1.0f);

						GL.Enable(EnableCap.PolygonOffsetFill);
						GL.PolygonMode(MaterialFace.Front, PolygonMode.Fill);
						GL.Color4(spherePart.material.diffuseColor);
						spherePart.Display();
						GL.Disable(EnableCap.PolygonOffsetFill);

						GL.PolygonMode(MaterialFace.Front, PolygonMode.Line);
						spherePart.DisplayWireframe();
						GL.PolygonMode(MaterialFace.Front, PolygonMode.Fill);
					}

					GL.Disable(EnableCap.ColorMaterial);
				}
				else
				{
					GL.TexEnv(TextureEnvTarget.TextureEnv, TextureEnvParameter.TextureEnvMode, (int)TextureEnvMode.Replace);

					foreach (Object3D spherePart in panoramaSpheres)
						spherePart.Display();
				}
			}
		}

		void DrawMiniPlanet(int width, int height)
		{
			float ratio = (float)height / (float)width;
			float c = (float)Math.Cos(miniPlanetAngle);
			float s = (float)Math.Sin(miniPlanetAngle);

			Vector4 rot = miniPlanetScale * new Vector4(c, -s * ratio, s, c * ratio);

			GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);
			GL.MatrixMode(MatrixMode.Projection);
			GL.LoadIdentity();
			//Glu.Perspective((double)90, (double)width / (double)height, (double)0.1, (double)10.0);
			Glu.Ortho2D(0, width, 0, height);

			GL.MatrixMode(MatrixMode.Modelview);
			GL.LoadIdentity();

			GL.TexEnv(TextureEnvTarget.TextureEnv, TextureEnvParameter.TextureEnvMode, (int)TextureEnvMode.Replace);
			cubeMap.Begin();

			shader.Begin();
			GL.Uniform4(shader.GetUniformLocation("rot"), rot.X, rot.Y, rot.Z, rot.W);
			GL.Uniform1(shader.GetUniformLocation("zoom"), miniPlanetZoom);
			GL.Uniform1(shader.GetUniformLocation("power"), miniPlanetPower);
			GL.Uniform1(shader.GetUniformLocation("Cube"), 0);
			//GL.Arb.ActiveTexture(TextureUnit.Texture0);

			GL.Disable(EnableCap.DepthTest);
			GL.Begin(BeginMode.Triangles);
			GL.Vertex2(-1f, -1f);
			GL.Vertex2(3f, -1f);
			GL.Vertex2(-1f, 3f);
			GL.End();
			GL.Enable(EnableCap.DepthTest);

			shader.End();
			cubeMap.End();
		}

		void buttonSaveCubemap_Click(object sender, EventArgs e)
		{
			Cursor = Cursors.WaitCursor;

			GenerateAndSaveCubeMap();

			Cursor = Cursors.Default;

			LogMessage("Ready");
		}

		private void GenerateAndSaveCubeMap()
		{
			GenerateAndSaveCubeMapImage("posx.jpg", -90.0f, 0.0f);
			GenerateAndSaveCubeMapImage("negx.jpg", +90.0f, 0.0f);
			GenerateAndSaveCubeMapImage("posy.jpg", 180.0f, -90.0f);
			GenerateAndSaveCubeMapImage("negy.jpg", 180.0f, +90.0f);
			GenerateAndSaveCubeMapImage("posz.jpg", -180.0f, 0.0f);
			GenerateAndSaveCubeMapImage("negz.jpg", 0.0f, 0.0f);

			//GenerateAndSaveCubeMapImage("negx.jpg", -90.0f, 0.0f);
			//GenerateAndSaveCubeMapImage("posy.jpg", +90.0f, 0.0f);
			//GenerateAndSaveCubeMapImage("negy.jpg", 180.0f, -90.0f);
			//GenerateAndSaveCubeMapImage("posz.jpg", 180.0f, +90.0f);
			//GenerateAndSaveCubeMapImage("negz.jpg", -180.0f, 0.0f);
			//GenerateAndSaveCubeMapImage("posx.jpg", 0.0f, 0.0f);

			//GenerateAndSaveCubeMapImage("posx.jpg", panoramaTheta - 90.0f, panoramaPhi + 0.0f);
			//GenerateAndSaveCubeMapImage("negx.jpg", panoramaTheta + 90.0f, panoramaPhi + 0.0f);
			//GenerateAndSaveCubeMapImage("posy.jpg", panoramaTheta + 180.0f, panoramaPhi - 90.0f);
			//GenerateAndSaveCubeMapImage("negy.jpg", panoramaTheta + 180.0f, panoramaPhi + 90.0f);
			//GenerateAndSaveCubeMapImage("posz.jpg", panoramaTheta - 180.0f, panoramaPhi + 0.0f);
			//GenerateAndSaveCubeMapImage("negz.jpg", panoramaTheta + 0.0f, panoramaPhi + 0.0f);
		}

		void GenerateAndSaveCubeMapImage(string filename, float theta, float phi)
		{
			int maxCubeMapTextureSize;
			GL.GetInteger(GetPName.MaxCubeMapTextureSize, out maxCubeMapTextureSize);
			maxCubeMapTextureSize = Math.Min(2048, maxCubeMapTextureSize);

			RTTexture rtTexture = new RTTexture();
			rtTexture.Square = true;
			rtTexture.Init(maxCubeMapTextureSize, maxCubeMapTextureSize);
			rtTexture.BeginDraw();

			GL.MatrixMode(MatrixMode.Projection);
			GL.LoadIdentity();
			Glu.Perspective((double)90, (double)rtTexture.Width / (double)rtTexture.Height, (double)0.1, (double)10.0);

			GL.MatrixMode(MatrixMode.Modelview);
			DrawSphere(phi, theta);
			//GL.LoadIdentity();
			////Glu.LookAt(Vector3.Zero, -Vector3.UnitZ, Vector3.UnitY);

			//GL.Rotate(phi, Vector3.UnitX);
			//GL.Rotate(theta, Vector3.UnitY);

			////foreach (Object3D spherePart in panoramaSpheres)
			////    spherePart.Display();
			//if (DebugMaterial)
			//{
			//    GL.Enable(EnableCap.ColorMaterial);
			//    GL.TexEnv(TextureEnvTarget.TextureEnv, TextureEnvParameter.TextureEnvMode, (int)TextureEnvMode.Modulate);

			//    foreach (Object3D spherePart in panoramaSpheres)
			//    {
			//        GL.Color4(spherePart.material.diffuseColor);
			//        spherePart.Display();
			//        spherePart.DisplayWireframe();
			//    }

			//    GL.Disable(EnableCap.ColorMaterial);
			//}
			//else
			//{
			//    GL.TexEnv(TextureEnvTarget.TextureEnv, TextureEnvParameter.TextureEnvMode, (int)TextureEnvMode.Replace);

			//    foreach (Object3D spherePart in panoramaSpheres)
			//        spherePart.Display();
			//}

			Bitmap bitmap = rtTexture.GrabScreenshot();
			rtTexture.EndDraw();
			rtTexture.Release();

			LogMessage("Saving " + filename);
			try
			{
				bitmap.Save(Path.Combine(cacheFolder, filename), ImageFormat.Jpeg);
			}
			catch (Exception ex)
			{
				LogErrorMessage(string.Format("Cannot save the MiniPlanet image ('{0}').\r\n{1}", filename, ex.Message));
			}
			finally
			{
				if (bitmap != null)
					bitmap.Dispose();
			}
		}

		void ReleaseSpheres()
		{
			if (panoramaSpheres != null)
			{
				foreach (Object3D spherePart in panoramaSpheres)
					spherePart.material.Release();
			}
		}

		void buttonOK_Click(object sender, EventArgs e)
		{
			displayMode = DisplayMode.Panorama;

			Cursor = Cursors.WaitCursor;
			//http://maps.google.fr/maps?hl=fr&ie=UTF8&ll=48.390305,-4.486416&spn=0,359.995494&t=h&z=19&layer=c&cbll=48.390252,-4.486275&panoid=MXLJB6JGTFzqOSAr-pm-8g&cbp=12,23.835107765146837,,0,10.303571428571425
			//http://cbk0.google.com/cbk?output=tile&panoid=9aoJaoXlaD08JWgy2NIwiQ&zoom=3&x=6&y=1&cb_client=maps_sv
			PanoramaParameters panoramaParameters = GetPanoramaIdFromLatitudeLongitude(textBoxStreetViewURL.Text);
			if (!string.IsNullOrEmpty(panoramaParameters.PanoramaId))
			{
				try
				{
					ReleaseSpheres();

					CheckAndCreateCacheFolder();

					panoramaSpheres = CreateStreetViewSpheres(panoramaParameters, panoramaLoadingZoom);

					GenerateAndSaveCubeMap();

					ResizeControl();
					LogMessage("Ready");
				}
				catch (Exception ex)
				{
					Cursor = Cursors.Default;
					LogErrorMessage(string.Format("Cannot download the images: {0}", ex.Message));
				}
			}
			else
			{
				Cursor = Cursors.Default;
				LogErrorMessage(string.Format("The Url does not contain 'cbll' or 'panoid' parameter!"));
			}
			Cursor = Cursors.Default;
		}

		public class PanoramaParameters
		{
			public PanoramaParameters()
			{
				ImageWidth = 13312;
				ImageHeight = 6656;
				TileWidth = 512;
				TileHeight = 512;
				PanoramaId = null;
				NumZoomLevels = 5;
			}

			public int ImageWidth { get; set; }
			public int ImageHeight { get; set; }
			public int TileWidth { get; set; }
			public int TileHeight { get; set; }
			public string PanoramaId { get; set; }
			public int NumZoomLevels { get; set; }
		}

		PanoramaParameters GetPanoramaIdFromLatitudeLongitude(string streetViewURL)
		{
			PanoramaParameters result = new PanoramaParameters();

			string cbll = null;
			Match matchCBLL = Regex.Match(streetViewURL, @".*cbll=(?<cbll>[\d|\,|\-|\.]+)&?.*", RegexOptions.IgnoreCase | RegexOptions.Compiled);
			if (matchCBLL.Success)
				cbll = matchCBLL.Result("${cbll}");

			if (!string.IsNullOrEmpty(cbll))
			{
				try
				{
					string url = string.Format("http://maps.google.com/cbk?output=xml&ll={0}", cbll);
					LogMessage(string.Format("Download {0}", url));
					XmlTextReader reader = new XmlTextReader(url);
					XPathNavigator nav = new XPathDocument(reader).CreateNavigator();
					XPathNodeIterator nodeItor = nav.Select("panorama/data_properties");
					nodeItor.MoveNext();
					// image_width="2400" image_height="1200" tile_width="512" tile_height="512" pano_id="_DypCnJinAJMPPP4susYkg" num_zoom_levels="3" lat="36.111465" lng="-115.173094"
					result.PanoramaId = nodeItor.Current.GetAttribute("pano_id", "");
					result.ImageWidth = int.Parse(nodeItor.Current.GetAttribute("image_width", ""));
					result.ImageHeight = int.Parse(nodeItor.Current.GetAttribute("image_height", ""));
					result.TileWidth = int.Parse(nodeItor.Current.GetAttribute("tile_width", ""));
					result.TileHeight = int.Parse(nodeItor.Current.GetAttribute("tile_height", ""));
					result.NumZoomLevels = int.Parse(nodeItor.Current.GetAttribute("num_zoom_levels", ""));
					reader.Close();
				}
				catch (Exception ex)
				{
					//throw new Exception(string.Format("Cannot load the xml ll description from Google StreetView server.\r\n{0}", ex.Message), ex);
					LogMessage(string.Format("Cannot load the xml ll description from Google StreetView server.\r\n{0}", ex.Message));
				}
			}

			if (string.IsNullOrEmpty(result.PanoramaId))
			{
				Match matchPanoId = Regex.Match(streetViewURL, @".*&panoid=(?<panoid>[\w|\-]+)&.*", RegexOptions.IgnoreCase | RegexOptions.Compiled);
				if (matchPanoId.Success)
					result.PanoramaId = matchPanoId.Result("${panoid}");
			}

			return result;
		}

		void CheckAndCreateCacheFolder()
		{
			if (!Directory.Exists(cacheFolder))
				Directory.CreateDirectory(cacheFolder);
		}

		void checkBoxDebug_CheckedChanged(object sender, EventArgs e)
		{
			DebugMaterial = checkBoxDebug.Checked;
		}

		void buttonLoadCache_Click(object sender, EventArgs e)
		{
			displayMode = DisplayMode.Panorama;
			try
			{
				ReleaseSpheres();

				panoramaSpheres = CreateStreetViewSpheres(new PanoramaParameters(), panoramaLoadingZoom);

				GenerateAndSaveCubeMap();

				ResizeControl();
				LogMessage("Ready");
			}
			catch (Exception ex)
			{
				Cursor = Cursors.Default;
				LogErrorMessage(string.Format("Cannot load images: {0}", ex.Message));
			}
		}

		void buttonResetView_Click(object sender, EventArgs e)
		{
			if (displayMode == DisplayMode.Panorama)
			{
				panoramaTheta = 0.0f;
				panoramaPhi = 0.0f;
				panoramaZoom = 1.0f;
			}
			else if (displayMode == DisplayMode.MiniPlanet)
			{
				miniPlanetZoom = 0.1f;
				miniPlanetScale = 5.0f;
				miniPlanetPower = 1.2f;
				miniPlanetAngle = 0;
			}

			ResizeControl();
		}

		void comboBoxZoom_SelectedIndexChanged(object sender, EventArgs e)
		{
			panoramaLoadingZoom = comboBoxZoom.SelectedIndex;
		}

		void LogMessage(string text)
		{
			textBoxLog.AppendText("\r\n");
			textBoxLog.AppendText(text);
			Application.DoEvents();
		}

		void LogErrorMessage(string text)
		{
			MessageBox.Show(text);
			LogMessage(text);
		}


		Point oldMousePosition;

		void glControl_MouseWheel(object sender, MouseEventArgs e)
		{
			if (e.Delta != 0)
			{
				if (displayMode == DisplayMode.Panorama)
				{
					if (e.Delta > 0)
						panoramaZoom /= 1.250f;
					else if (e.Delta < 0)
						panoramaZoom *= 1.250f;

					if (panoramaZoom > 1.56249988)
						panoramaZoom = 1.56249988f;

					if (panoramaZoom < 0.005902958)
						panoramaZoom = 0.005902958f;

					ResizeControl();
				}
				else if (displayMode == DisplayMode.MiniPlanet)
				{
					if (ModifierKeys == Keys.Shift)
					{
						if (e.Delta > 0)
							miniPlanetPower -= 0.01f;
						else if (e.Delta < 0)
							miniPlanetPower += 0.01f;

						if (miniPlanetPower > 10.0f)
							miniPlanetPower = 10.0f;

						if (miniPlanetPower < -10.0f)
							miniPlanetPower = -10.0f;
					}
					else
					{
						if (e.Delta > 0)
							miniPlanetZoom -= 0.005f;
						else if (e.Delta < 0)
							miniPlanetZoom += 0.005f;

						if (miniPlanetZoom > 10.0f)
							miniPlanetZoom = 10.0f;

						if (miniPlanetZoom < -10.0f)
							miniPlanetZoom = -10.0f;
					}
					glControl.Invalidate();
				}
			}
		}

		private void glControl_MouseDown(object sender, MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left)
			{
				this.oldMousePosition = Cursor.Position;
				Cursor.Hide(); // bug mono Cursor.Hide();
			}
		}

		private void glControl_MouseMove(object sender, MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left && !this.oldMousePosition.IsEmpty && Cursor.Position != this.oldMousePosition)
			{
				Point newMousePosition = Cursor.Position;
				if (displayMode == DisplayMode.Panorama)
				{
					this.panoramaTheta += (float)(newMousePosition.X - this.oldMousePosition.X) / (float)Width * 360.0f;
					this.panoramaPhi += (float)(newMousePosition.Y - this.oldMousePosition.Y) / (float)Height * 180.0f;
				}
				else if (displayMode == DisplayMode.MiniPlanet)
				{
					this.miniPlanetAngle += (float)(newMousePosition.X - this.oldMousePosition.X) / (float)Width * 3.0f;
					this.miniPlanetScale += (float)(newMousePosition.Y - this.oldMousePosition.Y) / (float)Height * 30.0f;
				}

				Cursor.Position = this.oldMousePosition;
				glControl.Invalidate();
			}
		}

		private void glControl_MouseUp(object sender, MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Left)
			{
				Cursor.Show();
			}
		}

		private void glControl_KeyDown(object sender, KeyEventArgs e)
		{
			//float scale = 10.0f;
			//if (e.Shift)
			//    scale = 1.0f;

			//if (e.KeyCode == Keys.Left)
			//{
			//    panoramaTheta -= scale;
			//}
			//else if (e.KeyCode == Keys.Right)
			//{
			//    panoramaTheta += scale;
			//}
			//else if (e.KeyCode == Keys.Up)
			//{
			//    panoramaPhi -= scale;
			//}
			//else if (e.KeyCode == Keys.Down)
			//{
			//    panoramaPhi += scale;
			//}
			//else
			//    return;

			//glControl.Invalidate();
			//e.Handled = true;
		}

		private void buttonMiniPlanet_Click(object sender, EventArgs e)
		{
			string vertexShader = @"
varying vec2 pos;
uniform vec4 rot;

void main()
{
	gl_Position = gl_Vertex;

	pos.x = dot(gl_Vertex.xy, rot.xy);
	pos.y = dot(gl_Vertex.xy, rot.zw);
}
";
			string pixelShader;
			switch (comboBoxPlanetType.SelectedIndex)
			{
				case 1: // Keyhole world
					pixelShader = @"
varying vec2 pos;
uniform samplerCube Cube;
uniform float zoom;
uniform float power;

void main()
{
	vec3 texCoord;

	texCoord.zx = pos;
	texCoord.y = 1.0 - pow(dot(pos, pos), power) * zoom;
	
	gl_FragColor = textureCube(Cube, texCoord);
}
";
					break;
				case 2: // Keyhole world
					pixelShader = @"
varying vec2 pos;
uniform samplerCube Cube;
uniform float zoom;
uniform float power;

void main()
{
	vec3 texCoord;

	// Spin world
	float PI = 3.14159265358979323846264;
	float r = length(pos);
	float angle = acos(pos.y / r) * sign(-pos.x);
	float adjustedAngle = angle * 3.0;
	texCoord.x = r * cos(adjustedAngle);
	texCoord.z = r * sin(adjustedAngle);
	//float normalizeAngle = (PI + angle) / (2.0 * PI);
	//texCoord.y = (1.0 - 0.9 * normalizeAngle) * (pow(dot(pos, pos), power) * zoom - 1.0);
	texCoord.y = (pow(dot(pos, pos), power) * zoom - 1.0);
	
	
	gl_FragColor = textureCube(Cube, texCoord);
}
";
					break;
				default: // Standard MiniPlanet
					pixelShader = @"
varying vec2 pos;
uniform samplerCube Cube;
uniform float zoom;
uniform float power;

void main()
{
	vec3 texCoord;

	texCoord.xz = pos;
	texCoord.y = pow(dot(pos, pos), power) * zoom - 1.0;

	gl_FragColor = textureCube(Cube, texCoord);
}
";
					break;
			}


			shader.Load(vertexShader, pixelShader);
			if (shader.IsLoaded)
			{
				try
				{
					cubeMap.Folder = cacheFolder;
					cubeMap.Init();
					displayMode = DisplayMode.MiniPlanet;
					ResizeControl();
				}
				catch (Exception ex)
				{
					LogErrorMessage(string.Format("Cubemap initialization error: {0}", ex.Message));
				}
			}
			else
				LogErrorMessage("Sorry, you need a recent graphic card in order to generate the Mini-Planet.");
		}

		private void buttonSaveMiniPlanet_Click(object sender, EventArgs e)
		{
			SaveFileDialog saveFileDialog = new SaveFileDialog();
			saveFileDialog.Filter = "jpg files (*.jpg)|*.jpg|All files (*.*)|*.*";
			saveFileDialog.FilterIndex = 1;
			saveFileDialog.FileName = "MiniPlanet.jpg";
			saveFileDialog.RestoreDirectory = true;
			if (saveFileDialog.ShowDialog() == DialogResult.OK)
			{
				Cursor = Cursors.WaitCursor;
				RTTexture rtTexture = new RTTexture();
				rtTexture.Square = true;
				rtTexture.Init();
				rtTexture.BeginDraw();

				DrawMiniPlanet(rtTexture.Width, rtTexture.Height);
				Bitmap bitmap = rtTexture.GrabScreenshot();
				rtTexture.EndDraw();
				rtTexture.Release();

				LogMessage("Saving " + saveFileDialog.FileName);
				try
				{
					bitmap.Save(saveFileDialog.FileName, ImageFormat.Jpeg);
				}
				catch (Exception ex)
				{
					LogErrorMessage(string.Format("Cannot save the MiniPlanet image ('{0}').\r\n{1}", saveFileDialog.FileName, ex.Message));
				}
				finally
				{
					if (bitmap != null)
						bitmap.Dispose();
				}

				ResizeControl();

				LogMessage("Ready");
				Cursor = Cursors.Default;
			}
		}

		private void buttonAbout_Click(object sender, EventArgs e)
		{
			new AboutForm().ShowDialog(this);
		}

		private void MainForm_DragEnter(object sender, DragEventArgs e)
		{
			if (e.Data.GetDataPresent(DataFormats.Text) ||
				e.Data.GetDataPresent(DataFormats.FileDrop))
				e.Effect = DragDropEffects.Link;
			else
				e.Effect = DragDropEffects.None;
		}

		[DllImport("kernel32")]
		private static extern int GetPrivateProfileString(string section,
				 string key, string def, StringBuilder retVal, int size, string filePath);

		private void MainForm_DragDrop(object sender, DragEventArgs e)
		{
			if (e.Data.GetDataPresent(DataFormats.FileDrop))
			{
				string[] files = (string[])e.Data.GetData(DataFormats.FileDrop);
				if (files != null && files.Length > 0)
				{
					StringBuilder url = new StringBuilder(4096);
					int i = GetPrivateProfileString("InternetShortcut", "URL", "", url, url.Capacity, files[0]);
					SetUrl(url.ToString());
				}
			}
			else if (e.Data.GetDataPresent(DataFormats.Text))
			{
				SetUrl(e.Data.GetData(DataFormats.Text) as string);
			}
		}

		private void SetUrl(string url)
		{
			textBoxStreetViewURL.Text = url;
			int selectStartIndex = textBoxStreetViewURL.Text.IndexOf("panoid");
			int selectEndIndex = textBoxStreetViewURL.Text.IndexOf("&", selectStartIndex);
			textBoxStreetViewURL.SelectionStart = selectStartIndex;
			if (selectEndIndex > selectStartIndex)
				textBoxStreetViewURL.SelectionLength = selectEndIndex - selectStartIndex;

			textBoxStreetViewURL.Focus();
		}
	}
}
