//    Copyright (C) 2009  Regis COSNIER
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

using System;
using System.Drawing;
using System.Drawing.Imaging;
using OpenTK.Graphics;
using OpenTK.Math;

namespace StreetViewMapper
{
	public class Material
	{
		public string name = "";
		public enum MaterialType { Diffuse, Texture };
		public MaterialType materialType = MaterialType.Diffuse;
		public Vector4 diffuseColor = new Vector4(0.7f, 0.7f, 0.7f, 1.0f);
		public int textureName;
		public string textureFileName = "";

		public void Init()
		{
			Release();

			switch (materialType)
			{
				case MaterialType.Diffuse:
					materialType = MaterialType.Diffuse;
					break;
				case MaterialType.Texture:
					materialType = MaterialType.Texture;

					Bitmap bitmap = null;
					try
					{
						bitmap = new Bitmap(textureFileName);

						GL.GenTextures(1, out textureName);
						GL.Enable(EnableCap.Texture2D);
						GL.BindTexture(TextureTarget.Texture2D, textureName);

						BitmapData data = bitmap.LockBits(new Rectangle(0, 0, bitmap.Width, bitmap.Height),
							ImageLockMode.ReadOnly, System.Drawing.Imaging.PixelFormat.Format32bppArgb);
						GL.TexImage2D(TextureTarget.Texture2D, 0, PixelInternalFormat.Rgb, data.Width, data.Height, 0,
							OpenTK.Graphics.PixelFormat.Bgra,
							PixelType.UnsignedByte, data.Scan0);
						bitmap.UnlockBits(data);

						GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureWrapS, (int)TextureWrapMode.ClampToEdge);
						GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureWrapT, (int)TextureWrapMode.ClampToEdge);

						GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Linear);
						GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int)TextureMagFilter.Linear);

						GL.BindTexture(TextureTarget.Texture2D, 0);
						GL.Disable(EnableCap.Texture2D);
					}
					catch (Exception ex)
					{
						throw new Exception(string.Format("Cannot load the image ('{0}') from local cache.\r\n{1}", textureFileName, ex.Message), ex);
					}
					finally
					{
						if (bitmap != null)
							bitmap.Dispose();
					}
					break;
			}
		}

		public void Release()
		{
			if (textureName > 0)
			{
				GL.DeleteTextures(1, new int[] { this.textureName });
				this.textureName = 0;
			}
		}

		public void Begin()
		{
			switch (materialType)
			{
				case MaterialType.Diffuse:
					GL.Enable(EnableCap.ColorMaterial);
					GL.Color4(diffuseColor.X, diffuseColor.Y, diffuseColor.Z, diffuseColor.W);
					break;
				case MaterialType.Texture:
					GL.Enable(EnableCap.Texture2D);
					GL.BindTexture(TextureTarget.Texture2D, textureName);
					break;
			}
		}

		public void End()
		{
			switch (materialType)
			{
				case MaterialType.Diffuse:
					GL.Disable(EnableCap.ColorMaterial);
					break;
				case MaterialType.Texture:
					GL.BindTexture(TextureTarget.Texture2D, 0);
					GL.Disable(EnableCap.Texture2D);
					break;
			}
		}
	}
}
